#include "util.h"
#include "wm.h"
#include <time.h>
#include <json-c/json_object.h>
#include <json-c/json_tokener.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

static unsigned int node_id = 0;

TreeNode* wm_treenode_new(NodeType type, TreeNode *parent)
{
    DEBUG_PRINT("%s\n", __func__);
    TreeNode *node = calloc(sizeof(TreeNode), 1);

    node->parent = parent;
    node->type = type;
    node->children = wm_nodearray_new();

    node_id++;
    node->id = node_id;
    node->client = NULL;

    return node;
}

void wm_treenode_free(TreeNode *node)
{
    wm_nodearray_clear(node->children);
    wm_nodearray_free(node->children);
    free(node);
}

bool wm_treenode_is_empty(TreeNode *node)
{
    DEBUG_PRINT("%s\n", __func__);
    return (node->children->size == 0);
}

void wm_treenode_split_space(TreeNode *node, Rect *ret1, Rect *ret2)
{
    DEBUG_PRINT("%s\n", __func__);
    assert(node);
    assert(ret1);
    assert(ret2);

    switch (node->type) {
    case NODE_VERTICAL:
        ret1->w = node->pos.w / 2;
        ret1->h = node->pos.h;
        ret2->w = node->pos.w / 2;
        ret2->h = node->pos.h;

        ret1->x = node->pos.x;
        ret1->y = node->pos.y;

        ret2->x = node->pos.x + ret1->w;
        ret2->y = node->pos.y;
        return;
    case NODE_HORIZONTAL:
        ret1->w = node->pos.w;
        ret1->h = node->pos.h / 2;
        ret2->w = node->pos.w;
        ret2->h = node->pos.h / 2;

        ret1->x = node->pos.x;
        ret1->y = node->pos.y;

        ret2->x = node->pos.x;
        ret2->y = node->pos.y + ret1->h;
        return;
    default:
        fprintf(stderr, "wm: unhandled treenode type %d in %s, exiting.\n",
                node->type, __func__);
        abort();
    }
}

void wm_treenode_recalculate_space(TreeNode *node)
{
    assert(node);

    assert(node->children->size == 2);

    wm_treenode_split_space(node, &node->children->nodes[0]->pos,
                            &node->children->nodes[1]->pos);
}

void wm_treenode_swap(TreeNode *root, TreeNode *node1, TreeNode* node2)
{
    assert(root);
    assert(node1);
    assert(node2);

    assert(node1->type == NODE_CLIENT);
    assert(node2->type == NODE_CLIENT);

    Client *tmp = node1->client;
    node1->client = node2->client;
    node2->client = tmp;
}

void wm_node_type_to_str(NodeType type, char *buf, size_t bufsize)
{
    switch (type) {
    case NODE_CLIENT:
        strncpy(buf, "NODE_CLIENT", bufsize);
        return;
    case NODE_VERTICAL:
        strncpy(buf, "NODE_VERTICAL", bufsize);
        return;
    case NODE_HORIZONTAL:
        strncpy(buf, "NODE_HORIZONTAL", bufsize);
        return;
    case NODE_TAB:
        strncpy(buf, "NODE_TAB", bufsize);
        return;
    case NODE_COUNT:
        return;
    }
}

char* wm_treenode_to_str(TreeNode *node)
{
    assert(node);
    char type[128] = {0};
    wm_node_type_to_str(node->type, type, sizeof(type));

    json_object *jobj = json_object_new_object();

    json_object_object_add(jobj, "type", json_object_new_string(type));
    json_object_object_add(jobj, "parent", json_object_new_uint64((uint64_t)node->parent));
    json_object_object_add(jobj, "children", json_object_new_array());

    for (size_t i = 0; i < node->children->size; i++) {
        char *child_str = wm_treenode_to_str(node->children->nodes[i]);
        json_object_array_add(json_object_object_get(jobj, "children"),
            json_tokener_parse(child_str));
        free(child_str);
    }

    json_object *pos = json_object_new_object();
    json_object_object_add(pos, "x", json_object_new_int(node->pos.x));
    json_object_object_add(pos, "y", json_object_new_int(node->pos.y));
    json_object_object_add(pos, "w", json_object_new_int(node->pos.w));
    json_object_object_add(pos, "h", json_object_new_int(node->pos.h));
    json_object_object_add(jobj, "pos", pos);
    if (node->type == NODE_CLIENT && node->client) {
        json_object *client = json_object_new_object();
        json_object_object_add(client, "address", json_object_new_uint64((unsigned long)node->client));
        json_object_object_add(client, "name", json_object_new_string(node->client->name));
        json_object_object_add(client, "is_pid_set", json_object_new_boolean(node->client->is_pid_set));
        json_object_object_add(client, "pid", json_object_new_int(node->client->pid));
        json_object_object_add(jobj, "client", client);
    }
    json_object_object_add(jobj, "id", json_object_new_uint64((uint64_t)node->id));

    char *ret = strdup(json_object_to_json_string_ext(jobj, JSON_C_TO_STRING_PRETTY));

    json_object_put(jobj);

    return ret;
}

PtrArray wm_nonempty_workspaces_to_strptrarray(Wm *wm)
{
    PtrArray ret = wm_ptrarray_new();

    for (size_t i = 0; i < wm->smon->wscount; i++) {
        TreeNode *node = wm->smon->workspaces[i].tree;
        if (!wm_treenode_is_empty(node) || (node->type == NODE_CLIENT && node->client)) {
            WmWorkspaceToStrRet *ws_str_with_index = malloc(sizeof(WmWorkspaceToStrRet));
            *ws_str_with_index = (WmWorkspaceToStrRet) {
                .ws_index = i,
                .str = wm_treenode_to_str(wm->smon->workspaces[i].tree)
            };
            wm_ptrarray_push(&ret, ws_str_with_index);
        }
    }

    DEBUG_PRINT("%s returning %ld\n", __func__, ret.size);

    return ret;
}

void wm_log_state(Wm *wm, const char *prefixstr, const char* logfile)
{
    RETURN_IF_NULL(prefixstr);
    RETURN_IF_NULL(logfile);

    FILE *fptr = fopen(logfile, "a");
    if (!fptr) {
        fprintf(stderr, "wm: could not open log file %s\n", logfile);
        return;
    }

    PtrArray ws_str = wm_nonempty_workspaces_to_strptrarray(wm);

    char str[128] = {0};
    char prefix[WM_PREFIX_LEN + 32] = {0};
    snprintf(prefix, sizeof(prefix), "%lu_%s", time(NULL), prefixstr);

    json_object *log_entry_obj = json_object_new_object();
    json_object_object_add(log_entry_obj, prefix, json_object_new_object());

    for (size_t i = 0; i < ws_str.size; i++) {
        WmWorkspaceToStrRet *ws_with_index = (WmWorkspaceToStrRet*)ws_str.ptrs[i];
        snprintf(str, sizeof(str), "%ld", ws_with_index->ws_index);
        json_object_object_add(json_object_object_get(log_entry_obj, prefix), str, json_tokener_parse(ws_with_index->str));
        fprintf(fptr, "%s,\n", json_object_to_json_string_ext(log_entry_obj, JSON_C_TO_STRING_PRETTY));
        free(ws_with_index->str);
        free(ws_with_index);
    }

    json_object_put(log_entry_obj);
    fflush(fptr);
    fclose(fptr);
    wm_ptrarray_free(&ws_str);
}

TreeNode* wm_treenode_lmd(TreeNode *node)
{
    assert(node);

    if (node->children->size == 0) return node;

    return wm_treenode_lmd(node->children->nodes[0]);
}

NodeArray* wm_treenode_all_lmds(TreeNode *node)
{
    assert(node);

    NodeArray *postorder = wm_postorder_traversal(node);
    NodeArray *ret = wm_nodearray_new();

    for (size_t i = 0; i < postorder->size; i++) {
        TreeNode *lmd = wm_treenode_lmd(postorder->nodes[i]);

        wm_nodearray_push(ret, lmd);
    }

    wm_nodearray_free(postorder);
    return ret;
}

NodeArray* wm_postorder_traversal(TreeNode *tree)
{
    assert(tree);

    TreeNode *root = tree;

    typedef struct {
        TreeNode *node;
        size_t index;
    } NodeWithIndex;

    int root_index = 0;

    NodeArray* stack = wm_nodearray_new();
    NodeArray* ret = wm_nodearray_new();

    while (root != NULL || stack->size > 0) {
        if (root != NULL) {
            NodeWithIndex *n = malloc(sizeof(NodeWithIndex));
            n->node = root;
            n->index = root_index;
            wm_nodearray_push(stack, (void*)n);
            root_index = 0;

            if (root->children->size >= 1)
                root = root->children->nodes[0];
            else root = NULL;

            continue;
        }

        NodeWithIndex *nptr;
        wm_nodearray_pop(stack, (TreeNode**)(void**)&nptr);
        NodeWithIndex n = *nptr;
        free(nptr);
        wm_nodearray_push(ret, n.node);

        while (stack->size > 0 && n.index == ((NodeWithIndex**)stack->nodes)
               [stack->size-1]->node->children->size - 1) {
            wm_nodearray_pop(stack, (TreeNode**)(void**)&nptr);
            n = *nptr;
            free(nptr);

            wm_nodearray_push(ret, n.node);
        }

        if (stack->size > 0) {
            root = ((NodeWithIndex**)stack->nodes)[stack->size-1]->node
                   ->children->nodes[n.index + 1];
            root_index = n.index + 1;
        }
    }

    wm_nodearray_free(stack);

    return ret;
}

bool wm_is_treenode_keyroot(TreeNode *node)
{
    assert(node);
    if (!node->parent) return true;

    return wm_treenode_lmd(node) != wm_treenode_lmd(node->parent);
}

NodeArray* wm_treenode_all_keyroots(TreeNode *node)
{
    assert(node);

    NodeArray* ret = wm_nodearray_new();
    NodeArray* postorder = wm_postorder_traversal(node);

    for (size_t i = 0; i < postorder->size; i++) {
        TreeNode *node = postorder->nodes[i];
        if (wm_is_treenode_keyroot(node))
            wm_nodearray_push(ret, node);
    }

    wm_nodearray_free(postorder);
    return ret;
}

void wm_treenode_print(TreeNode *node)
{
    char *str = wm_treenode_to_str(node);
    DEBUG_PRINT("%s\n", str);
    free(str);
}

void wm_tree_to_DOT(TreeNode *root, const char *filename)
{
    DEBUG_PRINT("%s\n", __func__);

    FILE *fp = fopen(filename, "w");
    assert(fp);

    fprintf(fp, "digraph {\n");

    NodeArray *all_nodes = wm_all_nodes_to_array(root);
    NodeArray *printed_nodes = wm_nodearray_new();

    const int bufsize = 100;
    char buf1[bufsize];
    char buf2[bufsize];

    for (size_t i = 0; i < all_nodes->size; i++) {
        TreeNode *node = all_nodes->nodes[i];

        bool contains = false;
        for (size_t j = 0; j < printed_nodes->size; j++) {
            if (printed_nodes->nodes[j]->id == node->id)
                contains = true;
        }

        if (contains)
            continue;

        for (size_t j = 0; j < node->children->size; j++) {
            memset(buf1, 0, bufsize);
            memset(buf2, 0, bufsize);

            wm_node_type_to_str(node->type, buf1, bufsize);
            wm_node_type_to_str(node->children->nodes[j]->type, buf2, bufsize);

            fprintf(fp, "_%d_%s -> _%d_%s;\n", node->id, buf1,
                    node->children->nodes[j]->id, buf2);
        }
        wm_nodearray_push(printed_nodes, node);
    }

    fprintf(fp, "}");

    fflush(fp);
    fclose(fp);

    wm_nodearray_free(printed_nodes);
    wm_nodearray_free(all_nodes);
}

int wm_get_node_index(TreeNode *parent, unsigned int node_id)
{
    for (size_t i = 0; i < parent->children->size; i++) {
        if (node_id == parent->children->nodes[i]->id)
            return i;
    }

    return -1;
}

void wm_treenode_remove_node(Wm *wm, TreeNode *root, TreeNode *node)
{
    DEBUG_PRINT("%s\n", __func__);

    wm_tree_to_DOT(root, "wm_output.dot");

    assert(node);

    TreeNode *parent = node->parent;

    if (node->parent == NULL) {
        node->client = NULL;
        return;
    }

    for (size_t i = 0; i < parent->children->size; i++) {
        if (parent->children->nodes[i]->type == NODE_CLIENT &&
            parent->children->nodes[i] != node) {

            parent->type = NODE_CLIENT;
            parent->client = parent->children->nodes[i]->client;
            wm_treenode_free(node);
            wm_treenode_free(parent->children->nodes[i]);
            wm_nodearray_clear(parent->children);
            return;
        }
    }

    DEBUG_PRINT("%s: other_client was NULL!\n", __func__);

    TreeNode *sibling_node = wm_treenode_split_get_sibling(node);
    wm_treenode_free(node);

    // wm_nodearray_free(&node->children);

    // TODO
    size_t dock_y = 0;

    if (parent->parent == NULL) {
        // parent is root node
        DEBUG_PRINT("parent is root node!\n");
        sibling_node->parent = NULL;
        assert(node->type == NODE_CLIENT);
        node->client->ws->tree = sibling_node;

        node->client->ws->tree->pos = (Rect) {
            .x = wm->config.border_width, .y = wm->config.border_width + dock_y ,
            .w = node->client->ws->monitor->info.width - wm->config.border_width*2,
            .h = node->client->ws->monitor->info.height - wm->config.border_width*2 - dock_y,
        };

        assert(node->client->ws->tree->children->size == 2);
        wm_treenode_split_space(node->client->ws->tree, &node->client->ws->tree->children->nodes[0]->pos,
                                &node->client->ws->tree->children->nodes[1]->pos);
    } else {
        sibling_node->parent = parent->parent;
        *parent = *sibling_node;

        for (size_t i = 0; i < parent->children->size; i++) {
            parent->children->nodes[i]->parent = parent;
        }

        wm_treenode_recalculate_space(parent->parent);
        wm_treenode_recalculate_space(parent);
    }

    assert(node->client->ws->tree->children->size >= 1);
}

TreeNode* wm_treenode_split_get_sibling(TreeNode *node)
{
    for (size_t i = 0; i < node->parent->children->size; i++) {
        TreeNode *sibling_node = node->parent->children->nodes[i];
        if (sibling_node->id != node->id) {
            return sibling_node;
        }
    }

    return NULL;
}

TreeNode* wm_treenode_remove_client(Wm *wm, TreeNode *root, Client *client)
{
    DEBUG_PRINT("%s\n", __func__);
    TreeNode *client_node = wm_treenode_ptr_find_client_node(root, client);

    wm_tree_to_DOT(root, "wm_output.dot");

    if (client_node->parent == NULL) {
        client_node->client = NULL;
        return NULL;
    }

    TreeNode *node = client_node->parent;

    for (size_t i = 0; i < node->children->size; i++) {
        if (node->children->nodes[i]->type == NODE_CLIENT &&
            node->children->nodes[i]->client != client) {

            node->type = NODE_CLIENT;
            node->client = node->children->nodes[i]->client;
            wm_treenode_free(client_node);
            wm_treenode_free(node->children->nodes[i]);
            wm_nodearray_clear(node->children);
            return node;
        }
    }

    DEBUG_PRINT("%s: other_client was NULL!\n", __func__);

    TreeNode *sibling_node = wm_treenode_split_get_sibling(client_node);
    wm_treenode_free(client_node);

    // wm_nodearray_free(&node->children);

    // TODO
    size_t dock_y = 0;

    if (node->parent == NULL) {
        // parent is root node
        DEBUG_PRINT("parent is root node!\n");

        TreeNode old_root = *client->ws->tree;

        sibling_node->parent = NULL;
        *client->ws->tree = *sibling_node;

        wm_nodearray_free(old_root.children);
        free(sibling_node);

        client->ws->tree->pos = (Rect) {
            .x = wm->config.border_width, .y = wm->config.border_width + dock_y ,
            .w = client->ws->monitor->info.width - wm->config.border_width*2,
            .h = client->ws->monitor->info.height - wm->config.border_width*2 - dock_y,
        };

        assert(client->ws->tree->children->size == 2);

        client->ws->tree->children->nodes[0]->parent = client->ws->tree;
        client->ws->tree->children->nodes[1]->parent = client->ws->tree;
        wm_treenode_split_space(client->ws->tree, &client->ws->tree->children->nodes[0]->pos,
                                &client->ws->tree->children->nodes[1]->pos);
    } else {
        sibling_node->parent = node->parent;
        *node = *sibling_node;

        for (size_t i = 0; i < node->children->size; i++) {
            node->children->nodes[i]->parent = node;
        }

        wm_treenode_recalculate_space(node->parent);
        wm_treenode_recalculate_space(node);
    }

    assert(client->ws->tree->children->size >= 1);
    return node->children->nodes[0];
}

NodeArray* wm_nodearray_new()
{
    NodeArray *arr = calloc(1, sizeof(NodeArray));

    arr->capacity = 10;
    arr->nodes = calloc(arr->capacity, sizeof(TreeNode*));
    arr->size = 0;

    return arr;
}

void wm_nodearray_push(NodeArray *arr, TreeNode *node)
{
    assert(arr);

    arr->size++;

    if (arr->size >= arr->capacity) {
        TreeNode* temp = calloc(arr->capacity, sizeof(TreeNode*));
        memcpy(temp, arr->nodes, arr->capacity * sizeof(TreeNode*));
        free(arr->nodes);
        size_t old_capacity = arr->capacity;
        arr->capacity = old_capacity * 2;

        arr->nodes = calloc(arr->capacity, sizeof(TreeNode*));
        memcpy(arr->nodes, temp, old_capacity * sizeof(TreeNode*));

        free(temp);
        arr->nodes[arr->size - 1] = node;
        return;
    }

    arr->nodes[arr->size - 1] = node;
}

bool wm_nodearray_pop(NodeArray *arr, TreeNode **ret)
{
    DEBUG_PRINT("%s\n", __func__);
    assert(arr);
    assert(ret);

    if (arr->size == 0)
        return false;

    *ret = arr->nodes[arr->size - 1];
    arr->size--;
    return true;
}

bool wm_nodearray_pop_front(NodeArray *arr, TreeNode **ret)
{
    assert(arr);
    assert(ret);

    if (arr->size == 0)
        return false;

    *ret = arr->nodes[0];
    arr->size--;
    memmove(arr->nodes, arr->nodes+1, arr->size * sizeof(TreeNode*));

    return true;
}

void wm_nodearray_clear(NodeArray *arr)
{
    DEBUG_PRINT("%s\n", __func__);
    arr->size = 0;
    memset(arr->nodes, 0, arr->capacity * sizeof(TreeNode*));
}

bool wm_nodearray_remove(NodeArray *arr, size_t index)
{
    DEBUG_PRINT("%s\n", __func__);
    assert(arr);

    if (!arr)
        return false;

    if (index >= arr->size)
        return false;

    if (index == arr->size - 1) {
        arr->size--;
        return true;
    }

    memmove(arr->nodes + index, arr->nodes + index+1, arr->size-1 * sizeof(TreeNode*));

    return true;
}

void wm_nodearray_free(NodeArray *arr)
{
    assert(arr);

    arr->capacity = 0;
    arr->size = 0;
    free(arr->nodes);
    free(arr);
}

PtrArray wm_ptrarray_new()
{
    PtrArray arr;

    arr.capacity = 10;
    arr.ptrs = calloc(arr.capacity, sizeof(void*));
    arr.size = 0;

    return arr;
}

void wm_ptrarray_push(PtrArray *arr, void* ptr)
{
    assert(arr);

    arr->size++;

    if (arr->size >= arr->capacity) {
        void* temp = calloc(arr->capacity, sizeof(void*));
        memcpy(temp, arr->ptrs, arr->capacity * sizeof(void*));
        free(arr->ptrs);
        size_t old_capacity = arr->capacity;
        arr->capacity = old_capacity * 2;

        arr->ptrs = calloc(arr->capacity, sizeof(void*));
        memcpy(arr->ptrs, temp, old_capacity * sizeof(void*));

        free(temp);
        arr->ptrs[arr->size - 1] = ptr;
        return;
    }

    arr->ptrs[arr->size - 1] = ptr;
}

bool wm_ptrarray_pop(PtrArray *arr, void **ret)
{
    DEBUG_PRINT("%s\n", __func__);
    assert(arr);
    assert(ret);

    if (arr->size == 0)
        return false;

    *ret = arr->ptrs[arr->size - 1];
    arr->size--;
    return true;
}

bool wm_ptrarray_pop_front(PtrArray *arr, void **ret)
{
    assert(arr);
    assert(ret);

    if (arr->size == 0)
        return false;

    *ret = arr->ptrs[0];
    arr->size--;
    memmove(arr->ptrs, arr->ptrs+1, arr->size * sizeof(void*));

    return true;
}

bool wm_ptrarray_remove(PtrArray *arr, size_t index)
{
    DEBUG_PRINT("%s\n", __func__);
    assert(arr);

    if (!arr)
        return false;

    if (index >= arr->size)
        return false;

    if (index == arr->size - 1) {
        arr->size--;
        return true;
    }

    memmove(arr->ptrs + index, arr->ptrs + index+1, arr->size-1 * sizeof(void*));

    return true;
}

void wm_ptrarray_free(PtrArray *arr)
{
    // DEBUG_PRINT("%s\n", __func__);
    assert(arr);
    RETURN_IF_NULL(arr->ptrs);

    arr->capacity = 0;
    arr->size = 0;
    free(arr->ptrs);
    arr->ptrs = NULL;
}

TreeNode* wm_treenode_ptr_find_focused_client_node(TreeNode *root)
{
    DEBUG_PRINT("%s\n", __func__);
    assert(root);

    TreeNode *ret = NULL;

    NodeArray *visited = wm_nodearray_new();
    wm_nodearray_push(visited, root);

    PtrArray queue = wm_ptrarray_new();

    wm_ptrarray_push(&queue, root);

    while (queue.size > 0) {
        TreeNode *node;

        if (!wm_ptrarray_pop_front(&queue, (void**)&node))
            goto ret;

        if (node->type == NODE_CLIENT && node->client && node->client->focused) {
            ret = node;
            goto ret;
        }

        for (size_t i = 0; i < node->children->size; i++) {
            TreeNode *child_node = node->children->nodes[i];
            bool _visited = false;

            for (size_t j = 0; j < visited->size; j++) {
                if (visited->nodes[j]->id == child_node->id) {
                    _visited = true;
                    break;
                }
            }

            if (!_visited) {
                wm_nodearray_push(visited, child_node);
                wm_ptrarray_push(&queue, child_node);
            }
        }
    }

ret:
    wm_nodearray_free(visited);
    wm_ptrarray_free(&queue);
    return ret;
}

TreeNode* wm_treenode_ptr_find_client_node(TreeNode *root, Client *client)
{
    assert(root);
    assert(client);

    TreeNode *ret = NULL;
    NodeArray *client_nodes = wm_treenode_find_client_nodes(root);

    for (size_t i = 0; i < client_nodes->size; i++) {
        TreeNode *node = (TreeNode*)client_nodes->nodes[i];
        if (node->client == client) {
            ret = node;
            goto ret;
        }
    }

ret:
    wm_nodearray_free(client_nodes);
    return ret;
}

NodeArray *wm_treenode_find_client_nodes(TreeNode *root)
{
    assert(root);

    NodeArray *ret = wm_nodearray_new();
    NodeArray *all_nodes = wm_all_nodes_to_array(root);

    for (size_t i = 0; i < all_nodes->size; i++) {
        TreeNode *node = all_nodes->nodes[i];
        if (node->type == NODE_CLIENT && node->client)
            wm_nodearray_push(ret, node);
    }

    wm_nodearray_free(all_nodes);
    return ret;
}

NodeArray* wm_all_nodes_to_array(TreeNode *root)
{
    DEBUG_PRINT("%s\n", __func__);
    assert(root);

    NodeArray *ret = wm_nodearray_new();

    NodeArray *visited = wm_nodearray_new();
    wm_nodearray_push(visited, root);

    NodeArray *queue = wm_nodearray_new();

    wm_nodearray_push(queue, root);

    while (queue->size > 0) {
        TreeNode *node;

        if (!wm_nodearray_pop_front(queue, &node))
            goto ret;

        wm_nodearray_push(ret, node);

        for (size_t i = 0; i < node->children->size; i++) {
            TreeNode *child_node = node->children->nodes[i];
            bool _visited = false;

            for (size_t j = 0; j < visited->size; j++) {
                if (visited->nodes[j]->id == child_node->id) {
                    _visited = true;
                    break;
                }
            }

            if (!_visited) {
                wm_nodearray_push(visited, child_node);
                wm_nodearray_push(queue, child_node);
            }
        }
    }

ret:
    wm_nodearray_free(visited);
    wm_nodearray_free(queue);
    return ret;
}
